import decimal
import uuid
import datetime
import iso8601

from kaiso.attributes.bases import RelationshipReference, wraps_type
from kaiso.types import Attribute, DefaultableAttribute


class Outgoing(RelationshipReference):
    pass


class Incoming(RelationshipReference):
    pass


class PrimitiveTypeMixin(object):
    """ Add a basic `to_primitive` method, coercing value to
    `cls.primitive_type` unless it is `None`.
    """

    @classmethod
    def to_primitive(cls, value, for_db):
        if value is None:
            return None

        return cls.primitive_type(value)


@wraps_type(uuid.UUID)
class Uuid(PrimitiveTypeMixin, Attribute):
    primitive_type = str

    @property
    def default(self):
        return uuid.uuid4()

    @classmethod
    def to_python(cls, value):
        if value is None:
            return None
        return uuid.UUID(hex=value)


class Bool(PrimitiveTypeMixin, DefaultableAttribute):
    primitive_type = bool


class Integer(PrimitiveTypeMixin, DefaultableAttribute):
    primitive_type = int


class Float(PrimitiveTypeMixin, DefaultableAttribute):
    primitive_type = float


class String(PrimitiveTypeMixin, DefaultableAttribute):
    primitive_type = unicode


@wraps_type(tuple)
class Tuple(PrimitiveTypeMixin, DefaultableAttribute):
    primitive_type = list

    @classmethod
    def to_python(cls, value):
        if value is None:
            return None
        return tuple(value)


@wraps_type(decimal.Decimal)
class Decimal(PrimitiveTypeMixin, DefaultableAttribute):
    primitive_type = str

    @classmethod
    def to_python(cls, value):
        if value is None:
            return None
        try:
            return decimal.Decimal(value)
        except decimal.DecimalException as ex:
            # DecimalException doesn't inherit from ValueError
            raise ValueError(str(ex))


@wraps_type(datetime.datetime)
class DateTime(DefaultableAttribute):
    @classmethod
    def to_primitive(cls, value, for_db):
        if value is None:
            return None

        if isinstance(value, basestring):
            # make sure this is a valid date
            DateTime.to_python(value)
            return value

        try:
            return value.isoformat()
        except AttributeError as ex:
            raise ValueError(
                "{!r} is not a valid value for DateTime: {}".format(value, ex)
            )

    @classmethod
    def to_python(cls, value):
        if value is None:
            return None
        try:
            return iso8601.parse_date(value)
        except iso8601.ParseError as ex:
            # ParseError doesn't inherit from ValueError
            raise ValueError(str(ex))


class Choice(DefaultableAttribute):
    choices = Tuple(default=tuple(), required=True)

    def __init__(self, *choices, **kwargs):
        super(Choice, self).__init__(**kwargs)
        self.choices = choices



from datetime import datetime
import decimal
import pytest
from uuid import uuid4

import iso8601

from kaiso.attributes import Uuid, Integer, Decimal, DateTime
from kaiso.types import Entity


class TestDefaultToPrimitiveDefaultToPython(object):
    @pytest.fixture
    def cls(self):
        class Foo(Entity):
            bar = Integer()
        return Foo

    def test_correct_type(self, type_registry, cls):
        instance = cls(bar=1)
        data = type_registry.object_to_dict(instance, for_db=True)
        assert data['bar'] == 1
        obj = type_registry.dict_to_object(data)
        assert obj.bar == 1

    def test_coercable_type_valid_value(self, type_registry, cls):
        instance = cls(bar='1')
        data = type_registry.object_to_dict(instance, for_db=True)
        assert data['bar'] == 1
        obj = type_registry.dict_to_object(data)
        assert obj.bar == 1

    def test_coercable_type_invalid_value(self, type_registry, cls):
        instance = cls(bar='invalid')
        with pytest.raises(ValueError):
            type_registry.object_to_dict(instance, for_db=True)


class TestDefaultToPrimitiveCustomToPython(object):
    @pytest.fixture
    def cls(self):
        class Foo(Entity):
            bar = Decimal()
        return Foo

    def test_correct_type(self, type_registry, cls):
        instance = cls(bar=decimal.Decimal('2.1'))
        data = type_registry.object_to_dict(instance, for_db=True)
        assert data['bar'] == '2.1'
        obj = type_registry.dict_to_object(data)
        assert obj.bar == decimal.Decimal('2.1')

    def test_coercable_type_valid_value(self, type_registry, cls):
        instance = cls(bar='2.1')
        data = type_registry.object_to_dict(instance, for_db=True)
        assert data['bar'] == '2.1'
        obj = type_registry.dict_to_object(data)
        assert obj.bar == decimal.Decimal('2.1')

    def test_coercable_type_invalid_value(self, type_registry, cls):
        instance = cls(bar='invalid')
        with pytest.raises(ValueError) as ex:
            type_registry.object_to_dict(instance, for_db=True)
        assert "is not a valid value for" in str(ex)


class TestCustomToPrimitiveDefaultToPython(object):
    @pytest.fixture
    def cls(self):
        class Foo(Entity):
            bar = Uuid()
        return Foo

    def test_correct_type(self, type_registry, cls):
        uuid = uuid4()
        instance = cls(bar=uuid)
        data = type_registry.object_to_dict(instance, for_db=True)
        assert data['bar'] == str(uuid)
        obj = type_registry.dict_to_object(data)
        assert obj.bar == uuid

    def test_coercable_type_valid_value(self, type_registry, cls):
        uuid = uuid4()
        instance = cls(bar=str(uuid))
        data = type_registry.object_to_dict(instance, for_db=True)
        assert data['bar'] == str(uuid)
        obj = type_registry.dict_to_object(data)
        assert obj.bar == uuid

    def test_coercable_type_invalid_value(self, type_registry, cls):
        instance = cls(bar='invalid')
        with pytest.raises(ValueError):
            type_registry.object_to_dict(instance, for_db=True)


class TestCustomToPrimitiveCustomToPython(object):
    @pytest.fixture
    def cls(self):
        class Foo(Entity):
            bar = DateTime()
        return Foo

    def test_correct_type(self, type_registry, cls):
        value = datetime(2012, 1, 1, 2, 3, tzinfo=iso8601.iso8601.Utc())
        str_value = '2012-01-01T02:03:00+00:00'

        instance = cls(bar=value)
        data = type_registry.object_to_dict(instance, for_db=True)
        assert data['bar'] == str_value
        obj = type_registry.dict_to_object(data)
        assert obj.bar == value

    def test_coercable_type_valid_value(self, type_registry, cls):
        value = datetime(2012, 1, 1, 2, 3, tzinfo=iso8601.iso8601.Utc())
        str_value = '2012-01-01T02:03:00+00:00'

        instance = cls(bar=str_value)
        data = type_registry.object_to_dict(instance, for_db=True)
        assert data['bar'] == str_value
        obj = type_registry.dict_to_object(data)
        assert obj.bar == value

    def test_coercable_type_invalid_value(self, type_registry, cls):
        instance = cls(bar='invalid')
        with pytest.raises(ValueError):
            type_registry.object_to_dict(instance, for_db=True)

    def test_coercable_type_invalid_value_type(self, type_registry, cls):
        instance = cls(bar=object())
        with pytest.raises(ValueError):
            type_registry.object_to_dict(instance, for_db=True)



import pytest

from kaiso.exceptions import UnknownType
from kaiso.migration_helpers import get_type_registry_with_base_change
from kaiso.types import collector, Entity
from kaiso.attributes import Integer, String


def test_basic(manager):
    with collector() as collected:
        class A(Entity):
            pass

        class A2(A):
            pass

        class B(Entity):
            pass

        class B2(B):
            pass

        class AB(A, B):
            pass

    manager.save_collected_classes(collected)
    amended_registry = get_type_registry_with_base_change(
        manager, 'AB', ('A2', 'B2'))

    AmendedAB = amended_registry.get_class_by_id('AB')
    assert [c.__name__ for c in AmendedAB.mro()] == [
        'AB', 'A2', 'A', 'B2', 'B', 'Entity', 'AttributedBase',
        'Persistable', 'object'
    ]

    # check original registry is unchanged
    OriginalAB = manager.type_registry.get_class_by_id('AB')
    assert [c.__name__ for c in OriginalAB.mro()] == [
        'AB', 'A', 'B', 'Entity', 'AttributedBase',
        'Persistable', 'object'
    ]


def test_become_your_own_ancestor(manager):
    with collector() as collected:
        class A(Entity):
            pass

        class A2(A):
            pass

        class A3(A2):
            pass

    manager.save_collected_classes(collected)

    # become your own parent
    with pytest.raises(ValueError) as ex:
        get_type_registry_with_base_change(manager, 'A', ('A3',))
    assert "inheritance cycle" in str(ex)


def test_duplicate_base_class(manager):
    with collector() as collected:
        class A(Entity):
            pass

        class B(Entity):
            pass

        class C(A, B):
            pass

    manager.save_collected_classes(collected)

    with pytest.raises(ValueError) as ex:
        get_type_registry_with_base_change(manager, 'C', ('A', 'B', 'A'))
    assert "duplicate base class" in str(ex)


def test_move_down_the_hieararchy(manager):

    with collector() as collected:
        class A(Entity):
            pass

        class A2(A):
            pass

        class A3(A2):
            pass

        class A4(A3):
            pass

        class B(A):
            pass

        class C(B):
            pass

    manager.save_collected_classes(collected)

    get_type_registry_with_base_change(manager, 'B', ('A4',))


def test_bad_mro(manager):
    """
    from http://www.python.org/download/releases/2.3/mro/

     -----------
    |           |
    |    O      |
    |  /   \    |
     - X    Y  /
       |  / | /
       | /  |/
       A    B
       \   /
         ?
    """

    with collector() as collected:
        class X(Entity):
            pass

        class Y(Entity):
            pass

        class A(X, Y):
            pass

        class B(Y):  # to become B(Y, X)
            pass

        class AB(A, B):
            pass

    manager.save_collected_classes(collected)

    with pytest.raises(ValueError) as ex:
        get_type_registry_with_base_change(manager, 'B', ('Y', 'X'))
    assert "Cannot create a consistent method resolution" in str(ex)


def test_unknown_types(manager):
    with collector() as collected:
        class A(Entity):
            pass

    manager.save_collected_classes(collected)

    with pytest.raises(UnknownType):
        get_type_registry_with_base_change(manager, 'A', ['Invalid'])

    with pytest.raises(UnknownType):
        get_type_registry_with_base_change(manager, 'Invalid', ['A'])


def test_amended_indexes(manager):
    with collector() as collected:
        class A(Entity):
            id = Integer(unique=True)

        class B(Entity):
            code = String(unique=True)

    manager.save_collected_classes(collected)

    amended_registry = get_type_registry_with_base_change(
        manager, 'B', ('A',))

    # confirm that the "amended" indexes of B are now using both id and code
    amended_indexes = amended_registry.get_indexes_for_type(B)
    assert {(ind, key) for ind, key, _ in amended_indexes} == {
        ('a', 'id'), ('b', 'code')
    }


def test_amended_indexes_same_attr_name(manager):
    with collector() as collected:
        class A(Entity):
            id = Integer(unique=True)

        class B(Entity):
            id = String(unique=True)

        class C(A):
            pass

    manager.save_collected_classes(collected)

    amended_registry = get_type_registry_with_base_change(
        manager, 'C', ('A', 'B'))

    # confirm that the "amended" indexes of C are still just A.id
    amended_indexes = amended_registry.get_indexes_for_type(C)
    assert {(ind, key) for ind, key, _ in amended_indexes} == {
        ('a', 'id'),
    }
